package otrie

import (
	"fmt"
	"sort"
	"strings"

	"idiolink/go/netaddr"
)

var onesToMask = [9]byte{0, 128, 192, 224, 240, 248, 252, 254, 255}

func newCursor(start *oTrieNode) *cursor {
	initCap := 8 // Not too big, not too small?
	cur := &cursor{
		chain: make([]*oTrieNode, 1, initCap),
		coord: make([]int, 1, initCap),
	}
	cur.chain[0] = start
	return cur
}

type cursor struct {
	chain []*oTrieNode
	coord []int
}

func (c *cursor) String() (output string) {
	maskToOnes := func(m byte) int {
		for i, v := range onesToMask {
			if v == m {
				return i
			}
		}
		panic("BUG: *oTrieNode.Print(): maskToOnes()")
	}
	l := 0
	p := make([]string, 0, len(c.chain))
	for _, o := range c.chain[1:] {
		p = append(p, fmt.Sprintf("%d", o.octet))
		l += maskToOnes(o.mask)
	}
	output += strings.Join(p, ".")
	for _, cut := range c.chain[len(c.chain)-1].cuts {
		for i := 0; i < (cut - len(c.chain)); i++ {
			output += ".0"
		}
	}
	output += fmt.Sprintf("/%d", l)
	return
}

func (c *cursor) depth() int {
	return len(c.chain) - 1
}

func (c *cursor) descend() int {
	depth := c.depth()
	nextNodes := c.chain[depth].next
	if len(nextNodes) != 0 {
		nextNode := nextNodes[c.coord[depth]]
		c.coord = append(c.coord, 0)
		c.chain = append(c.chain, nextNode)
	}
	return c.depth()
}

func (c *cursor) ascend() int {
	depth := c.depth()
	c.coord[depth] = 0
	if depth == 0 {
		return depth
	}
	c.chain[depth] = nil
	c.coord = c.coord[0:depth]
	c.chain = c.chain[0:depth]
	c.coord[depth-1]++
	return c.depth()
}

func (c *cursor) curNode() *oTrieNode {
	return c.chain[c.depth()]
}

func (c *cursor) step() int {
	if c.coord[c.depth()] == len(c.curNode().next) {
		return c.ascend()
	}
	return c.descend()
}

func (c *cursor) stepMember() (depth int) {
	depth = c.depth()
	for len(c.curNode().cuts) == 0 {
		if c.coord[depth] == len(c.curNode().next) {
			depth = c.ascend()
			continue
		}
		depth = c.descend()
	}
	return
}

func contains(o1, m1, o2, m2 byte) bool {
	return m1 <= m2 && o1&m1 == o2&m1
}

func overlaps(o1, m1, o2, m2 byte) bool {
	return contains(o1, m1, o2, m2) ||
		contains(o2, m2, o1, m1)
}

func (c *cursor) reset() {
	for i := 1; i < len(c.chain); i++ {
		c.chain = nil
		c.coord[i] = 0
	}
	c.coord[0] = 0
}

/* Breadth-first search of OTrie */
func (c *cursor) search(
	p *netaddr.NetAddr,
	equiv ...func(byte, byte, byte, byte) bool,
) (match bool) {
	for c.coord[0] != len(c.chain[0].next) {
		if len(c.curNode().next) == 0 {
			return
		}
		octet := p.Address[c.depth()]
		mask :=
			onesToMask[byte(min(int(p.Length)-c.depth()*8, 8))]
		equivTest := overlaps
		if len(equiv) > 0 {
			equivTest = equiv[0]
		}
		matchTest := func(i int) bool {
			next := c.curNode().next[i]
			return !prefixLessThan(
				next.octet, next.mask, octet, mask,
			) || equivTest(
				next.octet, next.mask, octet, mask,
			)
		}
		n := len(c.curNode().next)
		c.coord[c.depth()] = sort.Search(n, matchTest)
		if c.coord[c.depth()] == n {
			return
		}
		next := c.curNode().next[c.coord[c.depth()]]
		if equivTest(next.octet, next.mask, octet, mask) &&
			next.hasCut(p.AddressSize()) {
			match = true
			return
		}
		_ = c.descend()
	}
	return
}

/* Depth-first iteration over OTrie */
func (c *cursor) iter() <-chan []*oTrieNode {
	ch := make(chan []*oTrieNode)
	go func() {
		defer close(ch)
		for c.coord[0] != len(c.chain[0].next) {
			if c.coord[c.depth()] == len(c.curNode().next) {
				if len(c.curNode().cuts) != 0 {
					tmp := make([]*oTrieNode, c.depth())
					copy(tmp, c.chain[1:])
					ch <- tmp
				}
			}
			_ = c.step()
		}
		c.reset()
	}()
	return ch
}

type OTrie interface {
	AddPrefix(*netaddr.NetAddr) error
	Contains(*netaddr.NetAddr) bool
	Overlaps(*netaddr.NetAddr) bool
	Print() string
}

func NewOTrie() OTrie {
	return &oTrieNode{}
}

func newOTrieNode(octet, mask byte) *oTrieNode {
	return &oTrieNode{
		octet: octet,
		mask:  mask,
	}
}

func min(a, b int) int {
	if a < b {
		return a
	}
	return b
}

/* Test against skewed total order on byte prefixes */
func prefixLessThan(o1, m1, o2, m2 byte) bool {
	f1 := int(o1) - int((o1^byte(255))&m1)
	f2 := int(o2) - int((o2^byte(255))&m2)
	return f1 < f2 && o1 != o2&m1
}

/* Every octet trie node represents the end of a prefix, but
* not all represented prefixes are intended to be members of
* the set. For example, when a /24 is added, antecedent
* nodes representing a /8 and /16 are also created:
*
*      /8 -> /16 -> /24
*
* The `cuts` field is used to determine which nodes
* represent members of the set and which do not. Nodes with
* an empty `cuts` do not represent members. A node with one
* cut represents one member whose total size in octets is
* given by the cut. A node with cuts {4, 128} simultaneously
* represents a 4-byte member and a 128-byte member, both of
* which share the same prefix octets and prefix length.
 */
type oTrieNode struct {
	octet byte
	mask  byte
	next  []*oTrieNode
	cuts  []int
}

func (n *oTrieNode) hasCut(cut int) bool {
	for _, c := range n.cuts {
		if c == cut {
			return true
		}
	}
	return false
}

func (n *oTrieNode) setCut(cut int) {
	if n.cuts == nil {
		n.cuts = make([]int, 0, 2)
	}
	if n.hasCut(cut) {
		return
	}
	n.cuts = append(n.cuts, cut)
}

func (root *oTrieNode) Print() (output string) {
	maskToOnes := func(m byte) int {
		for i, v := range onesToMask {
			if v == m {
				return i
			}
		}
		panic("BUG: *oTrieNode.Print(): maskToOnes()")
	}
	cur := newCursor(root)
	for chain := range cur.iter() {
		l := 0
		p := make([]string, 0, len(chain))
		for _, o := range chain {
			p = append(p, fmt.Sprintf("%d", o.octet))
			l += maskToOnes(o.mask)
		}
		output += strings.Join(p, ".")
		for _, cut := range chain[len(chain)-1].cuts {
			for i := 0; i < (cut - len(chain)); i++ {
				output += ".0"
			}
		}
		output += fmt.Sprintf("/%d\n", l)
	}
	return
}

func (root *oTrieNode) AddPrefix(p *netaddr.NetAddr) error {
	nodeLessThan :=
		func(cur *oTrieNode) func(int, int) bool {
			return func(i, j int) bool {
				return prefixLessThan(
					cur.next[i].octet, cur.next[i].mask,
					cur.next[j].octet, cur.next[j].mask,
				)
			}
		}
	curNode := root
	ones := int(p.Length)
	for _, o := range p.Address {
		if ones == 0 {
			break
		}
		mask := onesToMask[min(8, ones)]
		ones -= min(8, ones)
		octet := o & mask
		if curNode.next == nil {
			curNode.next = make([]*oTrieNode, 0, 64)
		}
		var nextNode *oTrieNode
		oIdx := sort.Search(len(curNode.next), func(i int) bool {
			return !prefixLessThan(
				curNode.next[i].octet,
				curNode.next[i].mask,
				octet,
				mask,
			)
		})
		switch {
		case oIdx != len(curNode.next) &&
			curNode.next[oIdx].octet == octet &&
			curNode.next[oIdx].mask == mask:
			nextNode = curNode.next[oIdx]
			if ones == 0 {
				nextNode.setCut(p.AddressSize())
			}
		default:
			nextNode = newOTrieNode(octet, mask)
			if ones == 0 {
				nextNode.setCut(p.AddressSize())
			}
			curNode.next = append(curNode.next, nextNode)
			if oIdx != len(curNode.next) {
				sort.Slice(curNode.next, nodeLessThan(curNode))
			}
		}
		curNode = nextNode
	}
	return nil
}

/* Test for overlap; after all, even if a given prefix does
* not contain a certain range, if the prefix is a subset of
* that range, then the range is relevant.
 */
func (root *oTrieNode) Overlaps(p *netaddr.NetAddr) bool {
	return newCursor(root).search(p)
}

func (root *oTrieNode) Contains(p *netaddr.NetAddr) bool {
	return newCursor(root).search(p, contains)
}
